==> frgal-3.10.1-py3-none-any.whl: frgal/tests/test_aio.py <==
"""Tests for aio module."""
import asyncio
from unittest.mock import Mock, call, patch, sentinel

import grpc
from google.protobuf.empty_pb2 import Empty
from grpc.aio import Channel
from testfixtures import LogCapture, SequenceComparison

from frgal.aio import AsyncGrpcClient, SyncGrpcProxy
from frgal.constants import GRPC_MAX_RECONNECT_BACKOFF_MS, GRPC_MIN_RECONNECT_BACKOFF_MS
from frgal.decoder import GrpcDecoder
from frgal.exceptions import ServiceNotFound, UnknownError
from frgal.utils import AsyncMock, AsyncTestClientMixin, make_awaitable

from .proto.example_pb2 import Reply

try:
    from unittest import IsolatedAsyncioTestCase as TestCase  # type: ignore
except ImportError:  # pragma: no cover
    from asynctest import TestCase  # type: ignore

try:
    from grpc._channel import _MultiThreadedRendezvous as _Rendezvous
except ImportError:  # pragma: no cover
    from grpc._channel import _Rendezvous


async def aiter(iterable):
    for x in iterable:
        yield x


class AsyncGrpcClientTest(TestCase):
    """Test AsyncGrpcClient."""

    def setUp(self):
        self.grpc_module_mock = Mock(spec=('coffeeStub', ))
        self.channel_mock = AsyncMock(spec=Channel)
        self.channel_mock._loop = asyncio.get_event_loop()
        self.default_options = (
            ('grpc.min_reconnect_backoff_ms', GRPC_MIN_RECONNECT_BACKOFF_MS),
            ('grpc.max_reconnect_backoff_ms', GRPC_MAX_RECONNECT_BACKOFF_MS),
        )

        with patch('grpc.aio.insecure_channel', return_value=self.channel_mock):
            self.client = AsyncGrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
            self.client.channel

        self.log_handler = LogCapture('frgal.aio')
        self.addCleanup(self.log_handler.uninstall)

    def test_connect_insecure(self):
        client = AsyncGrpcClient(sentinel.netloc)
        with patch('grpc.aio.insecure_channel', return_value=self.channel_mock) as mock:
            client.channel

        self.assertEqual(client.channel, self.channel_mock)
        self.assertEqual(mock.mock_calls, [call(sentinel.netloc, SequenceComparison(*self.default_options))])

    def test_connect_secure(self):
        client = AsyncGrpcClient(sentinel.netloc, credentials=sentinel.credentials)
        with patch('grpc.aio.secure_channel', return_value=self.channel_mock) as mock:
            client.channel

        self.assertEqual(client.channel, self.channel_mock)
        self.assertEqual(mock.mock_calls,
                         [call(sentinel.netloc, sentinel.credentials, SequenceComparison(*self.default_options))])

    def test_get_stub(self):
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        self.assertEqual(self.client.get_stub('coffee'), sentinel.coffee_stub)
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(self.channel_mock)])

    def test_get_stub_cache(self):
        # Test stubs are cached when using the same channel
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        # Get stub once
        self.client.get_stub('coffee')
        # Get stub twice
        self.assertEqual(self.client.get_stub('coffee'), sentinel.coffee_stub)
        # Check it was created only once
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(self.channel_mock)])

    def test_get_stub_unknown(self):
        with self.assertRaisesRegex(ServiceNotFound, 'Service teaStub could not be found'):
            self.client.get_stub('tea')

    def test_get_stub_closed_event_loop(self):
        # Test stubs are cached when using the same channel
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        # Get stub once
        self.client.get_stub('coffee')

        # Close event loop and prepare new channel
        self.client._channel._loop = asyncio.new_event_loop()  # type: ignore
        self.client._channel._loop.close()  # type: ignore
        channel_mock = AsyncMock(spec=Channel)
        channel_mock._loop = asyncio.get_event_loop()

        with patch('grpc.aio.insecure_channel', return_value=channel_mock) as channel:
            # Get stub second time
            self.assertEqual(self.client.get_stub('coffee'), sentinel.coffee_stub)

        self.assertEqual(channel.mock_calls, [call(sentinel.netloc, SequenceComparison(*self.default_options))])
        self.assertEqual(self.grpc_module_mock.mock_calls, [
            call.coffeeStub(self.channel_mock),
            call.coffeeStub(channel_mock),
        ])

    async def test_call(self):
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), 'Espresso!')

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned data { result: "Espresso!" }'),
        )

    async def test_call_closed_event_loop(self):
        self.client.channel._loop = asyncio.new_event_loop()
        self.client.channel._loop.close()
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)
        channel_mock = AsyncMock(spec=Channel)
        channel_mock._loop = asyncio.get_event_loop()

        with patch('grpc.aio.insecure_channel', return_value=channel_mock) as channel_mock:
            with patch('frgal.aio.get_call_id', return_value='XXXX'):
                self.assertEqual(await self.client.call('coffee', 'make', Empty()), 'Espresso!')

        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc, SequenceComparison(*self.default_options))])
        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned data { result: "Espresso!" }'),
        )

    async def test_call_empty(self):
        reply = Empty()
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), None)

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned '),
        )

    async def test_call_error(self):
        reply = Reply()
        reply.exception.tea_pot_error.CopyFrom(Reply.Exception.TeaPotError())
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with self.assertRaisesRegex(UnknownError, 'tea_pot_error'):
            await self.client.call('coffee', 'make', Empty())

    async def test_call_grpc_unavailable(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = [error, error, error, make_awaitable(reply)]

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), 'Espresso!')

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned data { result: "Espresso!" }'),
        )

    async def test_call_grpc_unavailable_too_many_times(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            with self.assertRaises(grpc.RpcError):
                await self.client.call('coffee', 'make', Empty(), max_retries=3)

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Retries were unsuccessful.'),
            ('frgal.aio', 'INFO',
             '[XXXX] make failed with <_MultiThreadedRendezvous of RPC that terminated with:\n'
             '\tstatus = StatusCode.UNAVAILABLE\n'
             '\tdetails = "Connection reset by peer"\n'
             '\tdebug_error_string = "None"\n'
             '>')
        )

    async def test_call_grpc_other_error(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = _Rendezvous(state, None, None, None)
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with patch('grpc.aio.insecure_channel', return_value=self.channel_mock) as channel_mock:
            with self.assertRaisesRegex(_Rendezvous, 'status = StatusCode.UNKNOWN'):
                await self.client.call('coffee', 'make', Empty())

        self.assertEqual(channel_mock.mock_calls, [])

    async def test_call_stream(self):
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = aiter([reply, reply])

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            async for response in self.client.call_stream('coffee', 'make', Empty()):  # pragma: no cover
                self.assertEqual(response, 'Espresso!')

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] stream make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] stream make returned 2 chunks'),
        )

    async def test_call_stream_error(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = _Rendezvous(state, None, None, None)
        stub_mock = self.grpc_module_mock.coffeeStub.return_value

        async def error_iterator():
            raise error
            yield None  # pragma: no cover

        stub_mock.make.return_value = error_iterator()

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            with self.assertRaisesRegex(_Rendezvous, 'status = StatusCode.UNKNOWN'):
                async for response in self.client.call_stream('coffee', 'make', Empty()):  # pragma: no cover
                    pass

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] stream make()'),
            ('frgal.aio', 'INFO',
             '[XXXX] stream make failed after 0 chunks with <_MultiThreadedRendezvous of RPC that terminated with:\n'
             '\tstatus = StatusCode.UNKNOWN\n'
             '\tdetails = "Unknown error"\n'
             '\tdebug_error_string = "None"\n'
             '>')
        )

    def test_decoder_default(self):
        client = AsyncGrpcClient(sentinel.netloc)

        self.assertEqual(type(client.decoder), GrpcDecoder)

    def test_decoder_custom(self):
        client = AsyncGrpcClient(sentinel.netloc, decoder=sentinel.decoder)

        self.assertEqual(client.decoder, sentinel.decoder)

    def test_decoder_class(self):
        class TestDecoder(GrpcDecoder):
            pass

        class TestClient(AsyncGrpcClient):
            decoder_cls = TestDecoder

        client = TestClient(sentinel.netloc)

        self.assertIsInstance(client.decoder, TestDecoder)

    def test_grpc_modules_default(self):
        client = AsyncGrpcClient(sentinel.netloc)

        self.assertEqual(client.grpc_modules, ())

    def test_grpc_modules_custom(self):
        client = AsyncGrpcClient(sentinel.netloc, grpc_modules=sentinel.grpc_modules)

        self.assertEqual(client.grpc_modules, sentinel.grpc_modules)

    def test_grpc_modules_class(self):
        class TestClient(AsyncGrpcClient):
            grpc_modules = sentinel.grpc_modules

        client = TestClient(sentinel.netloc)

        self.assertEqual(client.grpc_modules, sentinel.grpc_modules)

    def test_options_default(self):
        client = AsyncGrpcClient(sentinel.netloc)

        default_options = {
            'grpc.min_reconnect_backoff_ms': GRPC_MIN_RECONNECT_BACKOFF_MS,
            'grpc.max_reconnect_backoff_ms': GRPC_MAX_RECONNECT_BACKOFF_MS,
        }
        self.assertEqual(client.options, default_options)

    def test_options_custom(self):
        client = AsyncGrpcClient(sentinel.netloc, options={'key': sentinel.value})

        options = {
            'grpc.min_reconnect_backoff_ms': GRPC_MIN_RECONNECT_BACKOFF_MS,
            'grpc.max_reconnect_backoff_ms': GRPC_MAX_RECONNECT_BACKOFF_MS,
            'key': sentinel.value,
        }
        self.assertEqual(client.options, options)

    def test_options_override(self):
        client = AsyncGrpcClient(sentinel.netloc, options={'grpc.min_reconnect_backoff_ms': sentinel.value})

        options = {
            'grpc.min_reconnect_backoff_ms': sentinel.value,
            'grpc.max_reconnect_backoff_ms': GRPC_MAX_RECONNECT_BACKOFF_MS,
        }
        self.assertEqual(client.options, options)

    def test_options_class(self):
        class TestClient(AsyncGrpcClient):
            options = {'key': sentinel.value}

        client = TestClient(sentinel.netloc)

        self.assertEqual(client.options, {'key': sentinel.value})


class TestAsyncClient(AsyncTestClientMixin, AsyncGrpcClient):
    def get_coffee_sync(self) -> str:
        return 'Coffee'

    async def get_coffee_async(self) -> str:
        return 'AsyncCoffee'


class SyncGrpcProxyTest(TestCase):
    def setUp(self):
        self.log_handler = LogCapture('frgal.aio')
        self.addCleanup(self.log_handler.uninstall)

    def test_proxy_client(self):
        self.client = SyncGrpcProxy(TestAsyncClient(sentinel.netloc))
        self.assertEqual(self.client.netloc, sentinel.netloc)
        self.assertEqual(self.client.get_coffee_sync(), 'Coffee')
        self.assertEqual(self.client.get_coffee_async(), 'AsyncCoffee')
        with self.assertRaises(AttributeError):
            self.client.not_attr

    def test_proxy_client_cls(self):
        class SyncProxy(SyncGrpcProxy):
            client_cls = TestAsyncClient

        self.client = SyncProxy(sentinel.netloc)
        self.assertEqual(self.client.netloc, sentinel.netloc)
        self.assertEqual(self.client.get_coffee_sync(), 'Coffee')
        self.assertEqual(self.client.get_coffee_async(), 'AsyncCoffee')
        with self.assertRaises(AttributeError):
            self.client.not_attr

==> frgal-3.12.2-py3-none-any.whl: frgal/tests/test_aio.py <==
"""Tests for aio module."""
import asyncio
from unittest.mock import Mock, call, patch, sentinel

import grpc
from google.protobuf.empty_pb2 import Empty
from grpc.aio import Channel
from testfixtures import LogCapture, SequenceComparison

from frgal.aio import AsyncGrpcClient, SyncGrpcProxy
from frgal.constants import GRPC_MAX_RECONNECT_BACKOFF_MS, GRPC_MIN_RECONNECT_BACKOFF_MS
from frgal.decoder import GrpcDecoder
from frgal.exceptions import ServiceNotFound, UnknownError
from frgal.utils import AsyncMock, AsyncTestClientMixin, make_awaitable

from .proto.example_pb2 import Reply  # type: ignore[attr-defined]

try:
    from unittest import IsolatedAsyncioTestCase as TestCase  # type: ignore
except ImportError:  # pragma: no cover
    from asynctest import TestCase  # type: ignore

try:
    from grpc._channel import _MultiThreadedRendezvous as _Rendezvous
except ImportError:  # pragma: no cover
    from grpc._channel import _Rendezvous


async def aiter(iterable):
    for x in iterable:
        yield x


class AsyncGrpcClientTest(TestCase):
    """Test AsyncGrpcClient."""

    def setUp(self):
        self.grpc_module_mock = Mock(spec=('coffeeStub', ))
        self.channel_mock = AsyncMock(spec=Channel)
        self.channel_mock._loop = asyncio.get_event_loop()
        self.default_options = (
            ('grpc.min_reconnect_backoff_ms', GRPC_MIN_RECONNECT_BACKOFF_MS),
            ('grpc.max_reconnect_backoff_ms', GRPC_MAX_RECONNECT_BACKOFF_MS),
        )

        with patch('grpc.aio.insecure_channel', return_value=self.channel_mock):
            self.client = AsyncGrpcClient(sentinel.netloc, grpc_modules=[self.grpc_module_mock])
            self.client.channel

        self.log_handler = LogCapture('frgal.aio')
        self.addCleanup(self.log_handler.uninstall)

    def test_connect_insecure(self):
        client = AsyncGrpcClient(sentinel.netloc)
        with patch('grpc.aio.insecure_channel', return_value=self.channel_mock) as mock:
            client.channel

        self.assertEqual(client.channel, self.channel_mock)
        self.assertEqual(mock.mock_calls, [call(sentinel.netloc, SequenceComparison(*self.default_options))])

    def test_connect_secure(self):
        client = AsyncGrpcClient(sentinel.netloc, credentials=sentinel.credentials)
        with patch('grpc.aio.secure_channel', return_value=self.channel_mock) as mock:
            client.channel

        self.assertEqual(client.channel, self.channel_mock)
        self.assertEqual(mock.mock_calls,
                         [call(sentinel.netloc, sentinel.credentials, SequenceComparison(*self.default_options))])

    def test_get_stub(self):
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        self.assertEqual(self.client.get_stub('coffee'), sentinel.coffee_stub)
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(self.channel_mock)])

    def test_get_stub_cache(self):
        # Test stubs are cached when using the same channel
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        # Get stub once
        self.client.get_stub('coffee')
        # Get stub twice
        self.assertEqual(self.client.get_stub('coffee'), sentinel.coffee_stub)
        # Check it was created only once
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(self.channel_mock)])

    def test_get_stub_unknown(self):
        with self.assertRaisesRegex(ServiceNotFound, 'Service teaStub could not be found'):
            self.client.get_stub('tea')

    def test_get_stub_closed_event_loop(self):
        # Test stubs are cached when using the same channel
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        # Get stub once
        self.client.get_stub('coffee')

        # Close event loop and prepare new channel
        self.client._channel._loop = asyncio.new_event_loop()  # type: ignore
        self.client._channel._loop.close()  # type: ignore
        channel_mock = AsyncMock(spec=Channel)
        channel_mock._loop = asyncio.get_event_loop()

        with patch('grpc.aio.insecure_channel', return_value=channel_mock) as channel:
            # Get stub second time
            self.assertEqual(self.client.get_stub('coffee'), sentinel.coffee_stub)

        self.assertEqual(channel.mock_calls, [call(sentinel.netloc, SequenceComparison(*self.default_options))])
        self.assertEqual(self.grpc_module_mock.mock_calls, [
            call.coffeeStub(self.channel_mock),
            call.coffeeStub(channel_mock),
        ])

    async def test_call(self):
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), 'Espresso!')

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned data { result: "Espresso!" }'),
        )

    async def test_call_closed_event_loop(self):
        self.client.channel._loop = asyncio.new_event_loop()
        self.client.channel._loop.close()
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)
        channel_mock = AsyncMock(spec=Channel)
        channel_mock._loop = asyncio.get_event_loop()

        with patch('grpc.aio.insecure_channel', return_value=channel_mock) as channel_mock:
            with patch('frgal.aio.get_call_id', return_value='XXXX'):
                self.assertEqual(await self.client.call('coffee', 'make', Empty()), 'Espresso!')

        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc, SequenceComparison(*self.default_options))])
        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned data { result: "Espresso!" }'),
        )

    async def test_call_empty(self):
        reply = Empty()
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), None)

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned '),
        )

    async def test_call_error(self):
        reply = Reply()
        reply.exception.tea_pot_error.CopyFrom(Reply.Exception.TeaPotError())
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = make_awaitable(reply)

        with self.assertRaisesRegex(UnknownError, 'tea_pot_error'):
            await self.client.call('coffee', 'make', Empty())

    async def test_call_grpc_unavailable(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = [error, error, error, make_awaitable(reply)]

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            self.assertEqual(await self.client.call('coffee', 'make', Empty()), 'Espresso!')

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'DEBUG', '[XXXX] make returned data { result: "Espresso!" }'),
        )

    async def test_call_grpc_unavailable_too_many_times(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            with self.assertRaises(grpc.RpcError):
                await self.client.call('coffee', 'make', Empty(), max_retries=3)

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] make()'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Attempting retry.'),
            ('frgal.aio', 'INFO', '[XXXX] Connection failed. Retries were unsuccessful.'),
            ('frgal.aio', 'INFO',
             '[XXXX] make failed with <_MultiThreadedRendezvous of RPC that terminated with:\n'
             '\tstatus = StatusCode.UNAVAILABLE\n'
             '\tdetails = "Connection reset by peer"\n'
             '\tdebug_error_string = "None"\n'
             '>')
        )

    async def test_call_grpc_other_error(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = _Rendezvous(state, None, None, None)
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with patch('grpc.aio.insecure_channel', return_value=self.channel_mock) as channel_mock:
            with self.assertRaisesRegex(_Rendezvous, 'status = StatusCode.UNKNOWN'):
                await self.client.call('coffee', 'make', Empty())

        self.assertEqual(channel_mock.mock_calls, [])

    async def test_call_stream(self):
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = aiter([reply, reply])

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            async for response in self.client.call_stream('coffee', 'make', Empty()):  # pragma: no cover
                self.assertEqual(response, 'Espresso!')

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] stream make()'),
            ('frgal.aio', 'DEBUG', '[XXXX] stream make returned 2 chunks'),
        )

    async def test_call_stream_error(self):
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = _Rendezvous(state, None, None, None)
        stub_mock = self.grpc_module_mock.coffeeStub.return_value

        async def error_iterator():
            raise error
            yield None  # pragma: no cover

        stub_mock.make.return_value = error_iterator()

        with patch('frgal.aio.get_call_id', return_value='XXXX'):
            with self.assertRaisesRegex(_Rendezvous, 'status = StatusCode.UNKNOWN'):
                async for response in self.client.call_stream('coffee', 'make', Empty()):  # pragma: no cover
                    pass

        self.log_handler.check(
            ('frgal.aio', 'DEBUG', '[XXXX] stream make()'),
            ('frgal.aio', 'INFO',
             '[XXXX] stream make failed after 0 chunks with <_MultiThreadedRendezvous of RPC that terminated with:\n'
             '\tstatus = StatusCode.UNKNOWN\n'
             '\tdetails = "Unknown error"\n'
             '\tdebug_error_string = "None"\n'
             '>')
        )

    def test_decoder_default(self):
        client = AsyncGrpcClient(sentinel.netloc)

        self.assertEqual(type(client.decoder), GrpcDecoder)

    def test_decoder_custom(self):
        client = AsyncGrpcClient(sentinel.netloc, decoder=sentinel.decoder)

        self.assertEqual(client.decoder, sentinel.decoder)

    def test_decoder_class(self):
        class TestDecoder(GrpcDecoder):
            pass

        class TestClient(AsyncGrpcClient):
            decoder_cls = TestDecoder

        client = TestClient(sentinel.netloc)

        self.assertIsInstance(client.decoder, TestDecoder)

    def test_grpc_modules_default(self):
        client = AsyncGrpcClient(sentinel.netloc)

        self.assertEqual(client.grpc_modules, ())

    def test_grpc_modules_custom(self):
        client = AsyncGrpcClient(sentinel.netloc, grpc_modules=sentinel.grpc_modules)

        self.assertEqual(client.grpc_modules, sentinel.grpc_modules)

    def test_grpc_modules_class(self):
        class TestClient(AsyncGrpcClient):
            grpc_modules = sentinel.grpc_modules

        client = TestClient(sentinel.netloc)

        self.assertEqual(client.grpc_modules, sentinel.grpc_modules)

    def test_options_default(self):
        client = AsyncGrpcClient(sentinel.netloc)

        default_options = {
            'grpc.min_reconnect_backoff_ms': GRPC_MIN_RECONNECT_BACKOFF_MS,
            'grpc.max_reconnect_backoff_ms': GRPC_MAX_RECONNECT_BACKOFF_MS,
        }
        self.assertEqual(client.options, default_options)

    def test_options_custom(self):
        client = AsyncGrpcClient(sentinel.netloc, options={'key': sentinel.value})

        options = {
            'grpc.min_reconnect_backoff_ms': GRPC_MIN_RECONNECT_BACKOFF_MS,
            'grpc.max_reconnect_backoff_ms': GRPC_MAX_RECONNECT_BACKOFF_MS,
            'key': sentinel.value,
        }
        self.assertEqual(client.options, options)

    def test_options_override(self):
        client = AsyncGrpcClient(sentinel.netloc, options={'grpc.min_reconnect_backoff_ms': sentinel.value})

        options = {
            'grpc.min_reconnect_backoff_ms': sentinel.value,
            'grpc.max_reconnect_backoff_ms': GRPC_MAX_RECONNECT_BACKOFF_MS,
        }
        self.assertEqual(client.options, options)

    def test_options_class(self):
        class TestClient(AsyncGrpcClient):
            options = {'key': sentinel.value}

        client = TestClient(sentinel.netloc)

        self.assertEqual(client.options, {'key': sentinel.value})


class TestAsyncClient(AsyncTestClientMixin, AsyncGrpcClient):
    def get_coffee_sync(self) -> str:
        return 'Coffee'

    async def get_coffee_async(self) -> str:
        return 'AsyncCoffee'


class SyncGrpcProxyTest(TestCase):
    def setUp(self):
        self.log_handler = LogCapture('frgal.aio')
        self.addCleanup(self.log_handler.uninstall)

    def test_proxy_client(self):
        self.client = SyncGrpcProxy(TestAsyncClient(sentinel.netloc))
        self.assertEqual(self.client.netloc, sentinel.netloc)
        self.assertEqual(self.client.get_coffee_sync(), 'Coffee')
        self.assertEqual(self.client.get_coffee_async(), 'AsyncCoffee')
        with self.assertRaises(AttributeError):
            self.client.not_attr

    def test_proxy_client_cls(self):
        class SyncProxy(SyncGrpcProxy):
            client_cls = TestAsyncClient

        self.client = SyncProxy(sentinel.netloc)
        self.assertEqual(self.client.netloc, sentinel.netloc)
        self.assertEqual(self.client.get_coffee_sync(), 'Coffee')
        self.assertEqual(self.client.get_coffee_async(), 'AsyncCoffee')
        with self.assertRaises(AttributeError):
            self.client.not_attr
